@@ -4,7 +4,6 @@ import pprint
 import re
 
 from yea_wandb import backend
-from yea_wandb.mock_server import ParseCTX  # noqa: E402
 
 
 def fn_find(args, state):
@@ -27,6 +26,21 @@ def fn_find(args, state):
     return None
 
 
+def fn_count_regex(args, state):
+    assert isinstance(args, list)
+    assert len(args) == 3
+    var, data, pattern = args
+    err = []
+    d = parse_term(data, state=state, result=err)
+    assert not err
+
+    i = 0
+    for item in d:
+        if contains_regex([item], pattern):
+            i += 1
+    return i
+
+
 def fn_len(args, state):
     var = args
     assert isinstance(var, str)
@@ -57,12 +71,25 @@ def fn_sort(args, state):
     return list(sorted(lst))
 
 
+def fn_concat(args, state):
+    assert isinstance(args, list)
+    err = []
+    lst = []
+    for orig in args:
+        item = parse_term(orig, state=state, result=err)
+        lst.append(item)
+    assert not err
+    return "".join(lst)
+
+
 FNSTR = ":fn:"
 FNS = {
     "find": fn_find,
     "len": fn_len,
     "keys": fn_keys,
     "sort": fn_sort,
+    "count_regex": fn_count_regex,
+    "concat": fn_concat,
 }
 
 
@@ -112,6 +139,16 @@ def parse_term(v, state, result=None):
     return found
 
 
+def contains_regex(iterable, pattern):
+    """
+    v1: Iterable[Str], i.e Dict[Str --> Any], List[Str], etc.
+    v2: Str
+    """
+    for s in iterable:
+        if re.search(pattern, s):
+            return True
+
+
 OPSTR = ":op:"
 OPS = {
     "<": "__lt__",
@@ -123,6 +160,10 @@ OPS = {
     "contains": "__contains__",
     "not_contains": "__contains__",
 }
+OPS_FUNCS = {
+    "contains_regex": contains_regex,
+    "not_contains_regex": contains_regex,
+}
 
 
 def parse_expr(adict, state, result):
@@ -132,18 +173,22 @@ def parse_expr(adict, state, result):
     # might be an op
     if isinstance(k, str) and k.startswith(OPSTR) and isinstance(v, list):
         op = k[len(OPSTR) :]
-        opfunc = OPS.get(op)
+        opfunc = OPS.get(op) or OPS_FUNCS.get(op)
         if not opfunc:
             result.append("unknown op: {}".format(op))
             return not result
         assert len(v) == 2, "ops need 2 parameters"
         v1 = parse_term(v[0], state, result)
         v2 = parse_term(v[1], state, result)
-        f = getattr(v1, opfunc, None)
-        if f is None:
-            result.append("unimplemented op: {}".format(opfunc))
-            return not result
-        b = f(v2)
+        if op in OPS_FUNCS:
+            f = OPS_FUNCS.get(op)
+            b = f(v1, v2)
+        else:
+            f = getattr(v1, opfunc, None)
+            if f is None:
+                result.append("unimplemented op: {}".format(opfunc))
+                return not result
+            b = f(v2)
         # hack to invert result for not_contains
         if op.startswith("not_"):
             b = not (b)
@@ -169,7 +214,7 @@ class YeaWandbPlugin:
         self._name = "wandb"
 
     def monitors_init(self):
-        self._backend = backend.Backend(args=self._yc._args)
+        self._backend = backend.Backend(yc=self._yc, args=self._yc._args)
 
     def monitors_start(self):
         self._backend.start()
@@ -180,11 +225,13 @@ class YeaWandbPlugin:
     def monitors_reset(self):
         self._backend.reset()
 
-    def _get_backend_state(self, debug=False):
+    def _get_backend_state(self, debug=False, ytest=None):
         if not self._backend._server:
             # we are live
             return
 
+        from yea_wandb.mock_server import ParseCTX  # noqa: E402
+
         state = {}
         runs = []
 
@@ -197,6 +244,7 @@ class YeaWandbPlugin:
             parsed = ParseCTX(glob_ctx, run_id)
             ddict.update({run_id: parsed._debug()})
             run = {}
+            run["run_id"] = parsed.run_id
             run["config"] = parsed.config_user
             run["config_wandb"] = parsed.config_wandb
             run["summary"] = parsed.summary_user
@@ -204,6 +252,8 @@ class YeaWandbPlugin:
             run["telemetry"] = parsed.telemetry
             run["metrics"] = parsed.metrics
             run["exitcode"] = parsed.exit_code
+            run["files"] = parsed.files
+            run["output"] = parsed.output
             runs.append(run)
 
         state[":wandb:artifacts"] = glob_parsed.artifacts
@@ -211,6 +261,10 @@ class YeaWandbPlugin:
         # deprecate this
         state[":wandb:runs_len"] = len(runs)
 
+        # TODO(): yea should actually have its own checks, for now do it here
+        if ytest:
+            state[":yea:exit"] = ytest._retcode
+
         if debug:
             pp = pprint.PrettyPrinter(indent=2)
             # pp.pprint(ddict)
@@ -256,7 +310,7 @@ class YeaWandbPlugin:
         return result
 
     def test_check(self, t, debug=False):
-        state = self._get_backend_state(debug=debug)
+        state = self._get_backend_state(debug=debug, ytest=t)
         result_list = None
         if state:
             result_list = self._check_state(t, state)
